package com.huigou.bpmc.chart.application.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;

import com.huigou.bpmc.chart.application.BPMCFlowChartApplication;
import com.huigou.bpmc.chart.application.FlowChartPermissionsApplication;
import com.huigou.bpmc.chart.domain.model.ProcessAreas;
import com.huigou.bpmc.chart.domain.model.ProcessNode;
import com.huigou.bpmc.chart.domain.model.ProcessNodeFunction;
import com.huigou.bpmc.chart.domain.model.ProcessNodeLine;
import com.huigou.bpmc.chart.domain.model.ProcessNodeTemp;
import com.huigou.bpmc.chart.domain.query.ProcessNodeFunctionQueryRequest;
import com.huigou.bpmc.chart.repository.ProcessAreasRepository;
import com.huigou.bpmc.chart.repository.ProcessNodeFunctionRepository;
import com.huigou.bpmc.chart.repository.ProcessNodeLineRepository;
import com.huigou.bpmc.chart.repository.ProcessNodeRepository;
import com.huigou.bpmc.chart.repository.ProcessNodeTempRepository;
import com.huigou.bpmc.process.application.BPMCBusinessProcessApplication;
import com.huigou.bpmc.process.domain.model.BusinessProcess;
import com.huigou.bpmc.process.repository.BusinessProcessRepository;
import com.huigou.context.MessageSourceContext;
import com.huigou.context.Operator;
import com.huigou.context.ThreadLocalUtil;
import com.huigou.data.domain.model.CommonDomainConstants;
import com.huigou.data.query.model.QueryDescriptor;
import com.huigou.uasp.bmp.common.application.BaseApplication;
import com.huigou.uasp.bmp.opm.proxy.AccessApplicationProxy;
import com.huigou.util.ApplicationContextWrapper;
import com.huigou.util.ListUtil;
import com.huigou.util.StringUtil;

/**
 * 流程图绘制
 *
 * @author xx
 * @version V1.0
 * @ClassName: FlowNodeApplicationImpl
 * @date 2017-03-06 14:49
 */
@Service("bpmcFlowChartApplication")
public class BPMCFlowChartApplicationImpl extends BaseApplication implements BPMCFlowChartApplication {
    @Autowired
    private ProcessNodeRepository processNodeRepository;

    @Autowired
    private ProcessNodeTempRepository processNodeTempRepository;

    @Autowired
    private ProcessNodeLineRepository processNodeLineRepository;

    @Autowired
    private ProcessNodeFunctionRepository processNodeFunctionRepository;

    @Autowired
    private ProcessAreasRepository processAreasRepository;

    @Autowired
    private BusinessProcessRepository businessProcessRepository;

    @Autowired
    private AccessApplicationProxy accessApplication;

    @Autowired
    private BPMCBusinessProcessApplication bpmBusinessProcessApplication;

    @Override
    public BusinessProcess loadBusinessProcess(String businessProcessId) {
        BusinessProcess process = businessProcessRepository.findOne(businessProcessId);
        Assert.notNull(process, String.format(CommonDomainConstants.OBJECT_NOT_FOUND_BY_ID, businessProcessId, "流程"));
        return process;
    }

    @Override
    public BusinessProcess loadBusinessProcessByCode(String code) {
        Assert.hasText(code, CommonDomainConstants.ID_NOT_BLANK);
        List<BusinessProcess> datas = businessProcessRepository.findByCode(code);
        Assert.notEmpty(datas, String.format("未找到[%s]对应的流程记录!", code));
        return datas.get(0);
    }

    @Override
    public Map<String, Object> queryFlowNodesAndLines(String businessProcessId) {
        Assert.hasText(businessProcessId, CommonDomainConstants.ID_NOT_BLANK);
        Map<String, Object> map = new HashMap<String, Object>();
        List<ProcessNode> nodes = processNodeRepository.findByBusinessProcessIdOrderByObjectKindCodeAsc(businessProcessId);
        List<ProcessNodeLine> lines = processNodeLineRepository.findByBusinessProcessId(businessProcessId);
        List<ProcessAreas> areas = processAreasRepository.findByBusinessProcessId(businessProcessId);
        // 处理ID 页面图像显示的ID与数据库存储的ID不同
        for (ProcessNode node : nodes) {
            node.setId(node.getViewId());
        }
        map.put("nodes", nodes);
        map.put("lines", lines);
        map.put("areas", areas);
        return map;
    }

    @Override
    public ProcessNode loadFlowNodeByViewId(String businessProcessId, String viewId) {
        List<ProcessNode> nodes = this.processNodeRepository.findOneByBusinessProcessIdAndViewId(businessProcessId, viewId);
        if (nodes != null && nodes.size() > 0) {
            return nodes.get(0);
        }
        return null;
    }

    /**
     * 删除节点
     *
     * @param nodes
     *            节点列表
     * @return
     */
    private void deleteNode(List<ProcessNode> nodes) {
        if (nodes != null && nodes.size() > 0) {
            for (ProcessNode node : nodes) {
                processNodeRepository.delete(node);
            }
        }
    }

    /**
     * 删除临时节点
     * 
     * @param businessProcessId
     * @param nodes
     */
    private void deleteNodeTemp(String businessProcessId, List<ProcessNodeTemp> nodes) {
        if (nodes != null && nodes.size() > 0) {
            for (ProcessNodeTemp node : nodes) {
                this.processNodeFunctionRepository.deleteByViewId(node.getViewId());
            }
        }
        this.processNodeTempRepository.deleteByProcessId(businessProcessId);
    }

    @Override
    @Transactional
    public void deleteFlowChart(String businessProcessId) {
        Assert.hasText(businessProcessId, CommonDomainConstants.ID_NOT_BLANK);
        List<ProcessNode> dbNodes = processNodeRepository.findByBusinessProcessId(businessProcessId);
        this.deleteNode(dbNodes);
        processNodeLineRepository.deleteLineByProcessId(businessProcessId);
        List<ProcessNodeTemp> dbNodeTemps = processNodeTempRepository.findByBusinessProcessId(businessProcessId);
        this.deleteNodeTemp(businessProcessId, dbNodeTemps);
        processNodeFunctionRepository.deleteByProcessId(businessProcessId);
        processAreasRepository.deleteByProcessId(businessProcessId);
        BusinessProcess businessProcess = businessProcessRepository.findOne(businessProcessId);
        businessProcess.setIsFlowChart(0);
        businessProcessRepository.save(businessProcess);
    }

    /**
     * 按viewId 组合节点MAP
     * 
     * @param tenantId
     * @param businessProcessId
     * @return
     */
    private Map<String, ProcessNode> getProcessNodeMap(String businessProcessId) {
        List<ProcessNode> dbNodes = processNodeRepository.findByBusinessProcessId(businessProcessId);
        Map<String, ProcessNode> map = new HashMap<String, ProcessNode>(dbNodes.size());
        for (ProcessNode node : dbNodes) {
            map.put(node.getViewId(), node);
        }
        return map;
    }

    /**
     * 按viewId 组合临时节点MAP
     * 
     * @param tenantId
     * @param businessProcessId
     * @return
     */
    private Map<String, ProcessNodeTemp> getProcessNodeTempMap(String businessProcessId) {
        List<ProcessNodeTemp> dbNodes = this.processNodeTempRepository.findByBusinessProcessId(businessProcessId);
        Map<String, ProcessNodeTemp> map = new HashMap<String, ProcessNodeTemp>(dbNodes.size());
        for (ProcessNodeTemp node : dbNodes) {
            map.put(node.getViewId(), node);
        }
        return map;
    }

    @Override
    @Transactional
    public void saveFlowNodesAndLines(String businessProcessId, String chartDirection, List<ProcessNode> nodes, List<ProcessNodeLine> lines,
                                      List<ProcessAreas> areas) {
        Assert.hasText(businessProcessId, CommonDomainConstants.ID_NOT_BLANK);
        BusinessProcess businessProcess = businessProcessRepository.findOne(businessProcessId);
        Assert.notNull(businessProcess, String.format(CommonDomainConstants.OBJECT_NOT_FOUND_BY_ID, businessProcessId, "流程"));
        // 数据库中已存在的节点
        Map<String, ProcessNode> dbNodes = this.getProcessNodeMap(businessProcessId);
        // 临时节点
        Map<String, ProcessNodeTemp> dbNodeTemps = this.getProcessNodeTempMap(businessProcessId);
        if (nodes != null && nodes.size() > 0) {
            List<ProcessNode> saveNodes = new ArrayList<ProcessNode>(nodes.size());
            String viewId = null;
            for (ProcessNode node : nodes) {
                viewId = node.getViewId();
                ProcessNode oldNode = dbNodes.get(viewId);
                if (oldNode == null) {
                    oldNode = node;
                    oldNode.setBusinessProcessId(businessProcessId);
                } else {
                    oldNode.setCode(node.getCode());
                    oldNode.setName(node.getName());
                    oldNode.setObjectKindCode(node.getObjectKindCode());
                    oldNode.setXaxis(node.getXaxis());
                    oldNode.setYaxis(node.getYaxis());
                    oldNode.setRuleKind(node.getRuleKind());
                    oldNode.setInterfaceKind(node.getInterfaceKind());
                    oldNode.setQuoteId(node.getQuoteId());
                    dbNodes.put(viewId, null);
                }
                // 处理临时节点中存储的数据
                ProcessNodeTemp processNodeTemp = dbNodeTemps.get(viewId);
                if (processNodeTemp != null) {
                    oldNode.setLinkKindCodes(processNodeTemp.getLinkKindCodes());
                    oldNode.setOwnerId(processNodeTemp.getOwnerId());
                    oldNode.setOwnerName(processNodeTemp.getOwnerName());
                    oldNode.setRemark(processNodeTemp.getRemark());
                    oldNode.setFunctionCode(processNodeTemp.getFunctionCode());
                    oldNode.setEnName(processNodeTemp.getEnName());
                    oldNode.setNodeCode(processNodeTemp.getNodeCode());
                    dbNodeTemps.put(viewId, null);
                }
                Assert.hasText(oldNode.getViewId(), CommonDomainConstants.ID_NOT_BLANK);
                saveNodes.add(oldNode);
            }
            processNodeRepository.save(saveNodes);
            // 删除原有连线
            processNodeLineRepository.deleteLineByProcessId(businessProcessId);
            for (ProcessNodeLine line : lines) {
                line.setBusinessProcessId(businessProcessId);
            }
            processNodeLineRepository.save(lines);
            // 设置流程存在流程图
            businessProcess.setIsFlowChart(1);
        } else {
            businessProcess.setIsFlowChart(0);
        }
        // 修改流程图定义
        businessProcess.setChartDirection(chartDirection);
        businessProcessRepository.save(businessProcess);
        // 不存在的节点
        List<ProcessNode> delNodes = new ArrayList<ProcessNode>(dbNodes.size());
        for (String key : dbNodes.keySet()) {
            ProcessNode node = dbNodes.get(key);
            if (node != null) {
                delNodes.add(node);
            }
        }
        // 删除不存在的节点
        this.deleteNode(delNodes);
        // 不存在的临时节点
        List<ProcessNodeTemp> delNodeTemps = new ArrayList<ProcessNodeTemp>(dbNodeTemps.size());
        for (String key : dbNodeTemps.keySet()) {
            ProcessNodeTemp node = dbNodeTemps.get(key);
            if (node != null) {
                delNodeTemps.add(node);
            }
        }
        // 删除全部临时节点
        this.deleteNodeTemp(businessProcessId, delNodeTemps);
        // 处理区域信息保存
        processAreasRepository.deleteByProcessId(businessProcessId);
        if (areas != null && areas.size() > 0) {
            for (ProcessAreas area : areas) {
                area.setId(null);
                area.setBusinessProcessId(businessProcessId);
            }
            processAreasRepository.save(areas);
        }
    }

    @Override
    @Transactional
    public ProcessNodeTemp saveFlowNode(String businessProcessId, ProcessNodeTemp node) {
        Assert.hasText(businessProcessId, CommonDomainConstants.ID_NOT_BLANK);
        BusinessProcess businessProcess = businessProcessRepository.findOne(businessProcessId);
        Assert.notNull(businessProcess, String.format(CommonDomainConstants.OBJECT_NOT_FOUND_BY_ID, businessProcessId, "流程"));
        String viewId = node.getViewId();
        // 保存到临时记录 不影响正常节点保存
        List<ProcessNodeTemp> oldNodes = this.processNodeTempRepository.findOneByBusinessProcessIdAndViewId(businessProcessId, viewId);
        ProcessNodeTemp oldNode = null;
        if (oldNodes != null && oldNodes.size() > 0) {
            oldNode = oldNodes.get(0);
            node.setId(oldNode.getId());
            oldNode.fromEntity(node);
        } else {
            oldNode = node;
            oldNode.setId(null);
            oldNode.setBusinessProcessId(businessProcessId);
        }
        List<ProcessNodeFunction> funs = processNodeFunctionRepository.findByViewIdOrderBySequenceAsc(viewId);
        if (funs != null && funs.size() > 0) {
            Map<String, String> kindCodeMap = new LinkedHashMap<>(funs.size());
            for (ProcessNodeFunction fun : funs) {
                kindCodeMap.put(fun.getIcon(), "1");
            }
            oldNode.setLinkKindCodes(ListUtil.join(kindCodeMap.keySet(), ","));
        } else {
            oldNode.setLinkKindCodes("");
        }
        oldNode = processNodeTempRepository.save(oldNode);
        return oldNode;
    }

    @Override
    @Transactional
    public String saveProcessNodeFunction(ProcessNodeFunction processNodeFunction) {
        Assert.notNull(processNodeFunction, CommonDomainConstants.OBJECT_NOT_NULL);
        processNodeFunction.checkConstraints();
        if (processNodeFunction.isNew()) {
            processNodeFunction.setSequence(this.commonDomainService.getNextSequence(ProcessNodeFunction.class, "viewId", processNodeFunction.getViewId()));
        }
        processNodeFunction = (ProcessNodeFunction) this.commonDomainService.loadAndFillinProperties(processNodeFunction);
        processNodeFunction = processNodeFunctionRepository.save(processNodeFunction);
        return processNodeFunction.getId();
    }

    @Override
    public ProcessNodeFunction loadProcessNodeFunction(String id) {
        Assert.hasText(id, CommonDomainConstants.ID_NOT_BLANK);
        return processNodeFunctionRepository.findOne(id);
    }

    @Override
    @Transactional
    public void deleteProcessNodeFunction(List<String> ids) {
        List<ProcessNodeFunction> objs = processNodeFunctionRepository.findAll(ids);
        processNodeFunctionRepository.delete(objs);
    }

    @Override
    public Map<String, Object> queryProcessNodeFunction(ProcessNodeFunctionQueryRequest queryRequest) {
        queryRequest.checkConstraints();
        QueryDescriptor queryDescriptor = this.sqlExecutorDao.getQuery(QUERY_XML_FILE_PATH, "processNodeFunction");
        return this.sqlExecutorDao.executeQuery(queryDescriptor, queryRequest);
    }

    @Override
    @Transactional
    public void updateProcessNodeFunctionSequence(Map<String, Integer> map) {
        Assert.notNull(map, MessageSourceContext.getMessage("object.not.null"));
        this.commonDomainService.updateSequence(ProcessNodeFunction.class, map);
    }

    @Override
    @SuppressWarnings("unchecked")
    public Map<String, Object> queryFlowNodesAndLinesAndFunction(String businessProcessId) {
        BusinessProcess businessProcess = bpmBusinessProcessApplication.loadBusinessProcess(businessProcessId);
        Assert.notNull(businessProcess);
        Map<String, Object> map = this.queryFlowNodesAndLines(businessProcessId);
        Operator operator = ThreadLocalUtil.getOperator();
        // 查询定义的功能
        List<ProcessNodeFunction> tmps = processNodeFunctionRepository.findByBusinessProcessIdOrderBySequenceAsc(businessProcessId);
        FlowChartPermissionsApplication application = null;
        List<ProcessNodeFunction> objs = null;
        String userCode = businessProcess.getUserCode();
        if (StringUtil.isNotBlank(userCode)) {
            application = ApplicationContextWrapper.getBean(userCode, FlowChartPermissionsApplication.class);
        }
        // 权限校验
        if (application != null) {
            objs = application.checkProcessNodeFunction(tmps);
        } else {// 默认使用系统中的角色权限校验
            objs = this.checkProcessNodeFunction(tmps);
        }
        Map<String, List<ProcessNodeFunction>> funs = new HashMap<>();
        if (objs != null && objs.size() > 0) {
            List<ProcessNodeFunction> temp = null;
            for (ProcessNodeFunction obj : objs) {
                if (funs.containsKey(obj.getViewId())) {
                    temp = funs.get(obj.getViewId());
                } else {
                    temp = new ArrayList<ProcessNodeFunction>();
                }
                temp.add(obj);
                funs.put(obj.getViewId(), temp);
            }
        }
        map.put("functions", funs);
        // 节点权限验证
        List<ProcessNode> nodes = (List<ProcessNode>) map.get("nodes");
        if (nodes != null && nodes.size() > 0) {
            boolean flag = true;
            String functionCode = "";
            List<String> noPermissions = new ArrayList<>();
            for (ProcessNode node : nodes) {
                functionCode = node.getFunctionCode();
                if (StringUtil.isNotBlank(functionCode)) {
                    // 校验当前登录用户是否有功能权限
                    flag = accessApplication.checkPersonFunPermissions(operator.getUserId(), functionCode);
                    if (!flag) {
                        // 没有权限
                        noPermissions.add(node.getId());
                    }
                }
            }
            map.put("noPermissions", ListUtil.join(noPermissions, ","));
        }
        return map;
    }

    /**
     * 默认系统权限校验
     * 
     * @param nodeFunctions
     * @return
     */
    private List<ProcessNodeFunction> checkProcessNodeFunction(List<ProcessNodeFunction> nodeFunctions) {
        Operator operator = ThreadLocalUtil.getOperator();
        List<ProcessNodeFunction> objs = new ArrayList<>(nodeFunctions.size());
        if (nodeFunctions == null || nodeFunctions.size() == 0) {
            return objs;
        }
        // 判断功能是否有权限
        String sql = "select count(0) from sa_opfunction t where t.code=?";
        boolean flag = true;
        for (ProcessNodeFunction fun : nodeFunctions) {
            int count = this.sqlExecutorDao.queryToInt(sql, fun.getCode());
            if (count == 0) {
                fun.setIsFunction(0);
                objs.add(fun);
            } else {
                flag = accessApplication.checkPersonFunPermissions(operator.getUserId(), fun.getCode());
                if (flag) {
                    fun.setIsFunction(1);
                    objs.add(fun);
                }
            }
        }
        return objs;
    }

}
